import { HeadersInit, BodyInit } from "node-fetch";
/** Descriptive HTTP statuses that map to their status code */
export declare enum StatusCode {
    OK = 200,
    Created = 201,
    NoContent = 204,
    BadRequest = 400,
    Unauthorized = 401,
    Forbidden = 403,
    NotFound = 404,
    MethodNotAllowed = 405,
    Conflict = 409,
    ImATeapot = 418,
    TooManyRequests = 429,
    InternalServerError = 500,
    NotImplemented = 501
}
/** Checks if a status code is considered as successfull (currently anything 2**) */
export declare const statusCodeIsSuccessful: (code: number) => boolean;
interface BasicResponse {
    status: StatusCode;
    message?: string;
}
/** A simple response object that can contain an error message and any error object */
export interface ErrorResponse extends BasicResponse {
    errorMessage?: string;
    error?: any;
    isError: true;
}
/** Type guard to check if a response is of the type ErrorResponse */
export declare function isErrorResponse(resp: any): resp is ErrorResponse;
/** Type guard to check if a response is of the type StatusCodeError */
export declare function isStatusCodeError(err: any): err is StatusCodeError;
/** Error handler that returns an appropriate ErrorResponse
 * for errors of type StatusCodeError, ErrorResponse, string and Error.
 *
 * Throws the error if it is an unhandled type. */
export declare function handleError(e: any): ErrorResponse;
/** Returns an ErrorResponse object using the StatusCodeError that was passed in
 * @param error The StatusCodeError object to create the ErrorResponse with
*/
export declare function statusCodeErrorResponse(error: StatusCodeError): ErrorResponse;
/** Returns an ErrorResponse object using the Error that was passed in
 * @param error The error object to create the ErrorResponse with
 * @param status The desired status code of the ErrorResponse (defaults to 500 Internal Server Error)
 */
export declare function generalErrorResponse(error: Error, status?: StatusCode): ErrorResponse;
/** Returns an ErrorResponse object using the string that was passed in,
 * @param errorMessage The string error object to create the ErrorResponse with
 * @param status The desired status code of the ErrorResponse (defaults to 500 Internal Server Error)
 */
export declare function stringErrorResponse(errorMessage: string, status?: StatusCode): ErrorResponse;
/** Returns an ErrorResponse object using the provided params
 * @param status The desired status code (defaults to 500 Internal Server Error)
 * @param errorMessage A descriptive error message
 * @param error An Error object
 */
export declare function errorResponse(status?: StatusCode, errorMessage?: string, error?: any): ErrorResponse;
/** Returns an ErrorResponse object for bad requests */
export declare function badRequest(errorMessage: string, error?: any): ErrorResponse;
/** Returns an ErrorResponse object for unauthorized requests */
export declare function unauthorized(errorMessage: string, error?: any): ErrorResponse;
/** Returns an ErrorResponse object for not found requests */
export declare function notFound(errorMessage: string, error?: any): ErrorResponse;
/** Returns an ErrorResponse object for internal server error requests */
export declare function internalServerError(errorMessage: string, error?: any): ErrorResponse;
/** Returns an ErrorResponse object for forbidden requests */
export declare function forbidden(errorMessage: string, error?: any): ErrorResponse;
/** An error object used when throwing errors for non-successful network requests, to include their status code */
export declare class StatusCodeError extends Error {
    code: StatusCode;
    constructor(code: StatusCode, message: string);
}
/** Supported request methods */
export declare type RequestMethod = "GET" | "POST";
export interface RequestParameters {
    url: string;
    errorMessage?: string;
    headers?: HeadersInit;
    body?: BodyInit;
}
/** Gets content from a URL with the specified options, casting the return response as the provided type
 * @param options Request parameters to be used, including the url, optional request headers / body, and
 * an error message in the case that the request was not successful
 * @returns Object of the specified type, or an ErrorResponse if it was unsuccessful
 */
export declare function getContent<T>(options: RequestParameters): Promise<ErrorResponse | T>;
/** Posts content to a URL with the specified options, casting the return response as the provided type
 * @param options Request parameters to be used, including the url, optional request headers / body, and
 * an error message in the case that the request was not successful
 * @returns Object of the specified type, or an ErrorResponse if it was unsuccessful
 */
export declare function postContent<T>(options: RequestParameters): Promise<ErrorResponse | T>;
export {};
