import fetch, { HeadersInit, BodyInit } from "node-fetch";

/** Descriptive HTTP statuses that map to their status code */
export enum StatusCode {
    OK = 200,
    Created = 201,
    NoContent = 204,
    BadRequest = 400,
    Unauthorized = 401,
    Forbidden = 403,
    NotFound = 404,
    MethodNotAllowed = 405,
    Conflict = 409,
    ImATeapot = 418,
    TooManyRequests = 429,
    InternalServerError = 500,
    NotImplemented = 501
}

/** Checks if a status code is considered as successfull (currently anything 2**) */
export const statusCodeIsSuccessful = (code: number) => {
    switch (code) {
        case StatusCode.OK:
        case StatusCode.Created:
        case StatusCode.NoContent:
            return true;
        default:
            return false;
    }
}

interface BasicResponse {
    status: StatusCode;
    message?: string;
}

/** A simple response object that can contain an error message and any error object */
export interface ErrorResponse extends BasicResponse {
    errorMessage?: string;
    error?: any; // yeah idk what type this is going to be... ¯\_(ツ)_/¯
    isError: true;
}

/** Type guard to check if a response is of the type ErrorResponse */
export function isErrorResponse(resp: any): resp is ErrorResponse {
    const r = resp as Partial<ErrorResponse>
    return r.status != null && r.isError != null && r.isError && !statusCodeIsSuccessful(r.status);
}

/** Type guard to check if a response is of the type StatusCodeError */
export function isStatusCodeError(err: any): err is StatusCodeError {
    const r = err as Partial<StatusCodeError>
    return r.code != null && r.message != null;
}

/** Error handler that returns an appropriate ErrorResponse
 * for errors of type StatusCodeError, ErrorResponse, string and Error.
 * 
 * Throws the error if it is an unhandled type. */
export function handleError(e: any) {
    if (isStatusCodeError(e)) {
        return statusCodeErrorResponse(e);
    } else if (typeof e === 'string' || e instanceof String) {
        return stringErrorResponse(e.toString());
    } else if (e instanceof Error) {
        return generalErrorResponse(e);
    } else if (isErrorResponse(e)) {
        return e;
    } else {
        // if we can't figure out what what we are dealing with then
        // probably cannot recover...therefore, rethrow
        // Note to Self: Rethink my life choices and choose better libraries to use.
        // ^ person from the internet
        throw e;
    }
}

/** Returns an ErrorResponse object using the StatusCodeError that was passed in 
 * @param error The StatusCodeError object to create the ErrorResponse with
*/
export function statusCodeErrorResponse(error: StatusCodeError): ErrorResponse {
    return errorResponse(error.code, error.message);
}

/** Returns an ErrorResponse object using the Error that was passed in
 * @param error The error object to create the ErrorResponse with
 * @param status The desired status code of the ErrorResponse (defaults to 500 Internal Server Error)
 */
export function generalErrorResponse(error: Error, status = StatusCode.InternalServerError): ErrorResponse {
    return errorResponse(status, error.message, error);
}

/** Returns an ErrorResponse object using the string that was passed in,
 * @param errorMessage The string error object to create the ErrorResponse with
 * @param status The desired status code of the ErrorResponse (defaults to 500 Internal Server Error)
 */
export function stringErrorResponse(errorMessage: string, status = StatusCode.InternalServerError): ErrorResponse {
    return errorResponse(status, errorMessage);
}

/** Returns an ErrorResponse object using the provided params
 * @param status The desired status code (defaults to 500 Internal Server Error)
 * @param errorMessage A descriptive error message
 * @param error An Error object
 */
export function errorResponse(status = StatusCode.InternalServerError, errorMessage?: string, error?: any): ErrorResponse {
    return { status, errorMessage, error, isError: true };
}

/** Returns an ErrorResponse object for bad requests */
export function badRequest(errorMessage: string, error?: any) {
    return errorResponse(StatusCode.BadRequest, errorMessage, error);
}

/** Returns an ErrorResponse object for unauthorized requests */
export function unauthorized(errorMessage: string, error?: any) {
    return errorResponse(StatusCode.Unauthorized, errorMessage, error);
}

/** Returns an ErrorResponse object for not found requests */
export function notFound(errorMessage: string, error?: any) {
    return errorResponse(StatusCode.NotFound, errorMessage, error);
}

/** Returns an ErrorResponse object for internal server error requests */
export function internalServerError(errorMessage: string, error?: any) {
    return errorResponse(StatusCode.InternalServerError, errorMessage, error);
}

/** Returns an ErrorResponse object for forbidden requests */
export function forbidden(errorMessage: string, error?: any) {
    return errorResponse(StatusCode.Forbidden, errorMessage, error);
}

/** An error object used when throwing errors for non-successful network requests, to include their status code */
export class StatusCodeError extends Error {
    code: StatusCode;
    constructor(code: StatusCode, message: string) {
        super(message);
        this.code = code;
    }
}

/** Supported request methods */
export type RequestMethod = "GET" | "POST";

export interface RequestParameters {
    url: string;
    errorMessage?: string;
    headers?: HeadersInit;
    body?: BodyInit;
}

/** Gets content from a URL with the specified options, casting the return response as the provided type
 * @param options Request parameters to be used, including the url, optional request headers / body, and
 * an error message in the case that the request was not successful
 * @returns Object of the specified type, or an ErrorResponse if it was unsuccessful
 */
export async function getContent<T>(options: RequestParameters) {
    return makeRequest<T>("GET", options);
}

/** Posts content to a URL with the specified options, casting the return response as the provided type
 * @param options Request parameters to be used, including the url, optional request headers / body, and
 * an error message in the case that the request was not successful
 * @returns Object of the specified type, or an ErrorResponse if it was unsuccessful
 */
export async function postContent<T>(options: RequestParameters) {
    return makeRequest<T>("POST", options);
}

async function makeRequest<T>(method: RequestMethod, options: RequestParameters) {
    const { url, headers, body, errorMessage } = options;
    let error: any;
    let status = StatusCode.BadRequest;
    try {
        // try fetch the data
        const response = await fetch(url, {
            method,
            headers: {
                ...headers,
                Accept: "application/json",
                "Content-type": "application/json"
            }, body
        });

        // cast the json into the proper type if successful
        if (statusCodeIsSuccessful(response.status)) return await response.json() as T;
        else {
            // if there's an error, let's see if we can get it
            // if it doesn't work out, well, just move on
            try {
                error = await response.json() as any;
            } finally { }
        }
        // if it fails, we'll take note of the status code
        status = response.status;

    } catch (e) {
        // if we got an error, let's track that too
        console.error(e);
        error = e;
    }

    // we'll be here if either the status code wasn't 2**, or if some exception was thrown :/
    return errorResponse(
        status,
        errorMessage != null ? errorMessage : `Request to ${url} failed.`,
        error
    );
}
